using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Unclassified;

namespace DropExplorer
{
	public partial class MainForm : Form
	{
		private Hashtable dropData;
		private string monospaceFont;
		private object selectedData;
		private DelayedCall clipboardChangedDc;

		public MainForm()
		{
			InitializeComponent();

			foreach (FontFamily ff in FontFamily.Families)
			{
				if (ff.Name == "Consolas")
				{
					monospaceFont = ff.Name;
					break;
				}
			}
			if (monospaceFont == null)
			{
				monospaceFont = "Courier New";
			}

			Font = SystemFonts.MessageBoxFont;

			ContentPicture.Visible = false;
			clipboardChangedDc = DelayedCall.Create(ReadClipboard, 500);
		}

		private void MainForm_DragEnter(object sender, DragEventArgs e)
		{
			e.Effect = DragDropEffects.Copy;
		}

		[System.Runtime.InteropServices.DllImport("user32.dll")]
		public static extern bool SetForegroundWindow(IntPtr hWnd);

		private void ClearContent()
		{
			ContentWebBrowser.Visible = false;
			ContentWebBrowser.DocumentText = "";
			ContentText.Text = "";
			ContentText.Font = new Font(monospaceFont, 9);
			ContentText.Enabled = true;
			ContentPicture.Visible = false;
			ContentPicture.Image = null;
			selectedData = null;
		}

		private void ShowDataObject(IDataObject dataObject)
		{
			Debug.WriteLine("ShowDataObject()");
			ClearContent();
			FormatList.Items.Clear();
			dropData = new Hashtable();

			if (dataObject != null)
			{
				foreach (string format in dataObject.GetFormats())
				{
					object data = null;
					string displayName = format;
					try
					{
						//if (format == "FileContents")
						//{
						//    FileContentsDrop fcd = new FileContentsDrop(e);
						//    data = fcd.GetFiles();
						//}
						//else
						if (format == "EnhancedMetafile")
						{
							data = null;
							displayName += " (Unsupported by .NET)";
						}
						else if (dataObject.GetData(format) is MemoryStream)
						{
							displayName += " (MemoryStream)";
							MemoryStream stream = dataObject.GetData(format) as MemoryStream;
							data = stream.ToArray();
							//data = Encoding.Default.GetString(stream.ToArray());
							//data = Encoding.Unicode.GetString(stream.ToArray());

							//string str = "";
							//int i = stream.ReadByte();
							//int count = 0;
							//while (i != -1 && count < 1000)
							//{
							//    str += (char)i;
							//    count++;
							//    i = stream.ReadByte();
							//}
							//if (i != -1)
							//    str += "...";
							//data = str;
						}
						else
						{
							data = dataObject.GetData(format);
							if (data != null)
							{
								displayName += " (" + data.GetType().Name + ")";
							}
							else
							{
								displayName += " (null)";
							}
						}
					}
					catch (Exception ex)
					{
						MessageBox.Show(
							"An unexpected error occured while fetching the clipboard data.\n\n" + ex.ToString(),
							"Error",
							MessageBoxButtons.OK,
							MessageBoxIcon.Error);
					}
					dropData.Add(format, data);
					FormatList.Items.Add(new FormatItem(format, displayName));
				}
			}

			FormatList.Focus();
			if (FormatList.Items.Count > 0)
				FormatList.SelectedIndex = 0;
		}

		private void MainForm_DragDrop(object sender, DragEventArgs e)
		{
			// Focus() does nothing
			SetForegroundWindow(Handle);
			Debug.WriteLine("MainForm_DragDrop()");
			ShowDataObject(e.Data);
		}

		private void FormatList_SelectedIndexChanged(object sender, EventArgs e)
		{
			ClearContent();
			int index = FormatList.SelectedIndex;
			Debug.WriteLine("FormatList.SelectedIndex = " + index);
			if (index != -1)
			{
				FormatItem formatItem = FormatList.Items[index] as FormatItem;
				string format = formatItem.Format;
				object data = dropData[format];
				selectedData = data;

				if (format == "Rich Text Format" && data is string)
				{
					try
					{
						ContentText.Rtf = data as string;
					}
					catch (Exception ex)
					{
						MessageBox.Show(ex.ToString(), "RTF Exception");
					}
				}
				else if (format == "HTML Format" && data is string)
				{
					try
					{
						string strData = data as string;

						int startIndex = strData.IndexOf("<!DOCTYPE");
						if (startIndex == -1) startIndex = 0;
						strData = strData.Substring(startIndex);
						ContentWebBrowser.Visible = true;
						ContentWebBrowser.DocumentText = strData;
					}
					catch (Exception ex)
					{
						MessageBox.Show(ex.ToString(), "HTML Exception");
					}
				}
				else if (data is string)
				{
					ContentText.Text = data as string;
				}
				else if (data is string[])
				{
					ContentText.Text = "";
					foreach (string line in data as string[])
					{
						ContentText.Text += line + "\n";
					}
				}
				else if (data is byte[])
				{
					if (HexViewCheckbox.Checked)
					{
						StringBuilder sb = new StringBuilder();
						StringBuilder hex = new StringBuilder();
						StringBuilder raw = new StringBuilder();
						int pos = 0;
						foreach (byte b in data as byte[])
						{
							if (pos >= 1 * 1024 * 1024)
							{
								sb.Append("(...)");
								sb.Append(Environment.NewLine);
								break;
							}

							hex.Append(b.ToString("X2"));
							hex.Append(" ");
							if (b < 0x20)
								raw.Append(".");
							else
								raw.Append((char) b);

							if (raw.Length == 8)
							{
								hex.Append(" ");
								raw.Append(" ");
							}
							if (raw.Length == 16 + 1)
							{
								sb.Append(pos.ToString("X8"));
								sb.Append("  ");
								sb.Append(hex);
								sb.Append(" ");
								sb.Append(raw);
								sb.Append(Environment.NewLine);

								hex.Length = 0;
								raw.Length = 0;
								pos += 16;
							}
						}
						if (raw.Length > 0)
						{
							sb.Append(pos.ToString("X8"));
							sb.Append("  ");
							sb.Append(hex);
							if (raw.Length < 8)
							{
								sb.Append(" ");   // Space between both 8 bytes blocks, only one space character
								raw.Append(" ");   // Remember the space also on the "right side"
							}
							for (int i = raw.Length + 1; i <= 16 + 1; i++)
							{
								sb.Append("   ");
							}
							sb.Append(" ");
							sb.Append(raw);
						}

						ContentText.Text = sb.ToString();
					}
					else
					{
						if (UnicodeCheckbox.Checked)
							ContentText.Text = Encoding.Unicode.GetString(data as byte[]);
						else
							ContentText.Text = Encoding.Default.GetString(data as byte[]);
					}
				}
				else if (data is Bitmap)
				{
					ContentPicture.Visible = true;
					ContentPicture.Image = data as Bitmap;
				}
				else if (data == null)
				{
					ContentText.Text = "null";
					ContentText.Enabled = false;
				}
				else
				{
					ContentText.Text = data.GetType().ToString();
				}
			}
		}

		private void UpdateLink_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
		{
			UpdateCheck.Check("http://dev.unclassified.de/files/apps/dropexplorer/update.xml");
		}

		private void HomepageLink_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
		{
			Process.Start("http://dev.unclassified.de/apps/dropexplorer");
		}

		private void UnicodeCheckbox_CheckedChanged(object sender, EventArgs e)
		{
			FormatList_SelectedIndexChanged(null, null);   // Update display
		}

		private void HexViewCheckbox_CheckedChanged(object sender, EventArgs e)
		{
			FormatList_SelectedIndexChanged(null, null);   // Update display
		}

		private void ObserveClipboardCheck_CheckedChanged(object sender, EventArgs e)
		{
			if (ObserveClipboardCheck.Checked)
			{
				Debug.WriteLine("Adding ClipboardWatcher.Changed event handler");
				ClipboardWatcher.Changed += ClipboardWatcher_Changed;
			}
			else
			{
				Debug.WriteLine("Removing ClipboardWatcher.Changed event handler");
				ClipboardWatcher.Changed -= ClipboardWatcher_Changed;
			}
		}

		private void ClipboardWatcher_Changed(object sender, EventArgs e)
		{
			Debug.WriteLine("ClipboardWatcher_Changed()");
			// Start delayed to not block the clipboard viewer chain
			clipboardChangedDc.Reset();
		}

		private void ReadClipboard()
		{
			Debug.WriteLine("ReadClipboard()");
			GetClipboardButton_Click(null, null);
		}

		private void GetClipboardButton_Click(object sender, EventArgs e)
		{
			ShowDataObject(Clipboard.GetDataObject());
		}

		private void SaveButton_Click(object sender, EventArgs e)
		{
			SaveToFile();
		}

		private void MainForm_KeyDown(object sender, KeyEventArgs e)
		{
			if (e.KeyCode == Keys.S && e.Control && !e.Shift && !e.Alt)
			{
				e.Handled = true;
				e.SuppressKeyPress = true;
				DelayedCall.Start(SaveToFile, 1);
			}
			if (e.KeyCode == Keys.V && e.Control && !e.Shift && !e.Alt ||
				e.KeyCode == Keys.Insert && !e.Control && e.Shift && !e.Alt)
			{
				e.Handled = true;
				e.SuppressKeyPress = true;
				GetClipboardButton_Click(null, null);
			}
		}

		private void SaveToFile()
		{
			if (selectedData is string ||
				selectedData is string[] ||
				selectedData is byte[] ||
				selectedData is Bitmap)
			{
				SaveFileDialog dlg = new SaveFileDialog();
				dlg.CheckPathExists = true;
				dlg.OverwritePrompt = true;
				dlg.RestoreDirectory = true;
				dlg.Title = "Save selected data to a file";
				if (dlg.ShowDialog() == DialogResult.OK)
				{
					try
					{
						if (selectedData is string)
						{
							File.WriteAllText(dlg.FileName, selectedData as string, Encoding.UTF8);
						}
						else if (selectedData is string[])
						{
							File.WriteAllLines(dlg.FileName, selectedData as string[], Encoding.UTF8);
						}
						else if (selectedData is byte[])
						{
							File.WriteAllBytes(dlg.FileName, selectedData as byte[]);
						}
						else if (selectedData is Bitmap)
						{
							(selectedData as Bitmap).Save(dlg.FileName);
						}
					}
					catch (Exception ex)
					{
						MessageBox.Show(
							"An error occured while writing the file. " + ex.GetType().Name + ": " + ex.Message,
							"Error",
							MessageBoxButtons.OK,
							MessageBoxIcon.Error);
					}
				}
			}
			else
			{
				MessageBox.Show(
					"There is no data available to save to a file.",
					"Error",
					MessageBoxButtons.OK,
					MessageBoxIcon.Warning);
			}
		}
	}
}