using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using Unclassified;
using System.Drawing;
using Unclassified.Automation;

namespace Unclassified
{
	/// <summary>
	/// Provides mouse filtering features like hiding the cursor after a timeout of inactivity or dispatching
	/// mouse wheel input the the control below the mouse cursor.
	/// </summary>
	class MouseFilter : Component, IMessageFilter
	{
		// AutoHideCursor variables
		private bool autoHideCursor = false;
		private bool cursorHidden = false;
		private DelayedCall hideCursorTimer;
		private Point lastMousePoint = new Point(-1, -1);   // Invalid point, so any first seen mouse position is different

		// DispatchMouseWheel variables
		private bool dispatchMouseWheel = false;

		/// <summary>
		/// Fired when the mouse cursor was hidden.
		/// </summary>
		[Description("Fired when the mouse cursor was hidden.")]
		public event EventHandler MouseHidden;
		/// <summary>
		/// Fired when the mouse cursor was shown.
		/// </summary>
		[Description("Fired when the mouse cursor was shown.")]
		public event EventHandler MouseShown;

		public MouseFilter()
		{
			InitializeComponent();
		}

		public MouseFilter(IContainer container)
		{
			container.Add(this);
			InitializeComponent();
		}

		private void InitializeComponent()
		{
			hideCursorTimer = DelayedCall.Create(HideCursor, 2000);
			if (!DesignMode)
			{
				Application.AddMessageFilter(this);
			}
		}

		protected override void Dispose(bool disposing)
		{
			if (!DesignMode)
			{
				Application.RemoveMessageFilter(this);   // Removing it more than once is okay
			}
			base.Dispose(disposing);
		}

		/// <summary>
		/// Gets or sets a value indicating whether the mouse cursor should be hidden after it has not been moved for a while.
		/// </summary>
		[DefaultValue(false)]
		[Description("Hides the mouse cursor after a while of inactivity.")]
		public bool AutoHideCursor
		{
			get { return autoHideCursor; }
			set
			{
				autoHideCursor = value;
				if (!DesignMode)
				{
					if (value)
					{
						HideCursor();
						DelayedCall.Start(HideCursor, 100);
					}
					else
					{
						hideCursorTimer.Cancel();
						ShowCursor();
					}
				}
			}
		}

		/// <summary>
		/// Gets or sets a value indicating whether mouse wheel input should be dispatched to the control below the mouse cursor.
		/// </summary>
		/// <remarks>Normally, mouse wheel input always goes to the control which holds the keyboard
		/// input focus. But since the wheel is mounted to the mouse and not the keyboard, it should
		/// respect the mouse cursor's position more than the keyboard focus. Due to the way how Windows
		/// sends the corresponding messages however, you cannot scroll in an inactive window even with
		/// the cursor over it.</remarks>
		[DefaultValue(false)]
		[Description("Enables dispatching of mouse wheel input to the control below the mouse cursor.")]
		public bool DispatchMouseWheel
		{
			get { return dispatchMouseWheel; }
			set { dispatchMouseWheel = value; }
		}

		public bool PreFilterMessage(ref Message m)
		{
			if (m.Msg == WinApi.WM_MOUSEMOVE && autoHideCursor)
			{
				Point p = new Point(m.LParam.ToInt32());
				WinApi.ClientToScreen(m.HWnd, ref p);
				if (p != lastMousePoint)
				{
					lastMousePoint = p;
					ShowCursor();
					hideCursorTimer.Reset();
				}
				return false;   // Don't block this message
			}
			else if (m.Msg == WinApi.WM_MOUSEWHEEL && dispatchMouseWheel)
			{
				Point p = new Point(m.LParam.ToInt32());

				// First, see whether there's an open combobox list popup that must be handled differently
				WindowProperties prop = new WindowProperties();
				prop.ProcessId = new Window(m.HWnd).ProcessId;
				prop.ClassName = "ComboLBox";
				prop.Visible = true;
				Window popup;
				if (Window.TryFind(prop, out popup))
				{
					// Found a combobox popup window, check coordinates
					if (popup.Rectangle.Contains(p))
					{
						// Mouse cursor is over visible combo listbox popup, let the message pass
						return false;
					}
				}
				
				// Regular window surface: find the deepest control below the mouse cursor
				Control control = Form.ActiveForm;
				while (control != null)
				{
					Point clientPoint = p;
					WinApi.ScreenToClient(control.Handle, ref clientPoint);
					Control subControl = control.GetChildAtPoint(clientPoint, GetChildAtPointSkip.Invisible);
					if (subControl == null) break;
					//System.Diagnostics.Debug.WriteLine("Found control in " + control.Name + " at " + clientPoint + ": " + subControl.Name);
					control = subControl;
				}
				if (control != null)
				{
					if (control.Handle == m.HWnd) return false;   // The message already arrived for this control, let it pass

					// NOTE: Possible performance improvement: Flag the re-posted message so that we can
					//       recognise it faster when it re-arrives here and we can let it pass.
					//       Use either a very high delta value (HIWORD(wParam)) or one of the modifier
					//       keys MK_* (LOWORD(wParam)) or an unassigned bit of lParam.
					
					//System.Diagnostics.Debug.WriteLine("Re-posting message to " + control.Name);
					WinApi.PostMessage(control.Handle, WinApi.WM_MOUSEWHEEL, (uint) m.WParam.ToInt32(), (uint) m.LParam.ToInt32());

					// If IMessageModifyAndFilter wasn't System.Windows.Form's internal, we could just
					// implement that interface, modify the message and let it pass changed. But so we
					// need to block the message and send a new one that we're going to let pass then.
				}
				return true;   // Block this message
			}
			return false;   // Don't block this message
		}

		/// <summary>
		/// Hides the cursor if it's currently shown.
		/// </summary>
		/// <remarks>Since calls to <c>Cursor.Hide</c> and <c>Cursor.Show</c> must be symmetrical,
		/// these functions keep their own counter to make sure they're not called too often.</remarks>
		private void HideCursor()
		{
			if (!cursorHidden)
			{
				Cursor.Hide();
				cursorHidden = true;
				if (MouseHidden != null) MouseHidden(this, EventArgs.Empty);
			}
		}

		/// <summary>
		/// Shows the cursor if it's currently hidden.
		/// </summary>
		/// <remarks>Since calls to <c>Cursor.Hide</c> and <c>Cursor.Show</c> must be symmetrical,
		/// these functions keep their own counter to make sure they're not called too often.</remarks>
		private void ShowCursor()
		{
			if (cursorHidden)
			{
				Cursor.Show();
				cursorHidden = false;
				if (MouseShown != null) MouseShown(this, EventArgs.Empty);
			}
		}
	}
}
